from multiprocessing import Process, Manager
import signal
from flask import Response
from flask import Flask
from flask import render_template
import argparse
import os
import datetime
import time
import cv2
import gc

# initialize a flask object
app = Flask(__name__)

# constants
faceCascade = cv2.CascadeClassifier(cv2.data.haarcascades + 'haarcascade_frontalface_default.xml')

# global variables
camera_feed = None
web_feed = None

# Handle interrupt
def signal_handle(_signal, frame):
    print('Stopping the Jobs.')

# Write data to the shared buffer stack:
def camera_input(stack, cam, top: int) -> None:
    """
         :param cam: camera parameters
         :param stack: Manager.list object
         :param top: buffer stack capacity
    :return: None
    """
    print('Process to read camera: %s' % os.getpid())
	# signal trap to handle keyboard interrupt
    signal.signal(signal.SIGINT, signal_handle)

    cap = cv2.VideoCapture(cam)
    time.sleep(2.0)
	
    while True:
        _, img = cap.read()
        if _:
            stack.append(img)
            # Clear the buffer stack every time it reaches a certain capacity
            # Use the gc library to manually clean up memory garbage to prevent memory overflow
            if len(stack) >= top:
                del stack[:]
                gc.collect()

# Generate frame from the buffer stack:
def web_stream(stack) -> None:
    print('Process to read: %s' % os.getpid())
	# signal trap to handle keyboard interrupt
    #signal.signal(signal.SIGINT, signal_handle)
    while True:
        # if len(stack) != 0:
        try:
            frame = stack.pop()

            # grab the current timestamp and draw it on the frame
            timestamp = datetime.datetime.now()
            cv2.putText(frame, timestamp.strftime(
                "%A %d %B %Y %I:%M:%S%p"), (10, frame.shape[0] - 10),
                cv2.FONT_HERSHEY_SIMPLEX, 0.35, (0, 0, 255), 1)
			
            # encode the frame in JPEG format
            (flag, encodedImage) = cv2.imencode(".jpg", frame)

            # ensure the frame was successfully encoded
            if not flag:
                continue

            # yield the output frame in the byte format
            yield(b'--frame\r\n' b'Content-Type: image/jpeg\r\n\r\n' + 
                bytearray(encodedImage) + b'\r\n')
        except IndexError as e:
            # print(e)
            pass

# Process input images and write processed images to the shared buffer stack:
def video_processing(stack_in, stack_out, top: int) -> None:
    """
         :param stack_in: Manager.list buffer to get images
         :param stack_out: Manager.list buffer to write processed images
         :param top: buffer stack capacity
    :return: None
    """
    print('Process to process video: %s' % os.getpid())
	# signal trap to handle keyboard interrupt
    signal.signal(signal.SIGINT, signal_handle)

    while True:
        # if len(stack_in) != 0:
        try:
            frame = stack_in.pop()
            
			# Do the necessary video processing
			# E.g. face detection
            # draw face boxes
            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            faces = faceCascade.detectMultiScale(
                gray,     
                scaleFactor=1.2,
                minNeighbors=5,     
                minSize=(20, 20)
            )
            for (x,y,w,h) in faces:
                cv2.rectangle(frame,(x,y),(x+w,y+h),(255,0,0),2)
                roi_gray = gray[y:y+h, x:x+w]
                roi_color = frame[y:y+h, x:x+w]
				
			# scale image
            h, w = frame.shape[:2]
            frame = cv2.resize(frame, (400, int(400*h/w)))
			
            stack_out.append(frame)
            # Clear the buffer stack every time it reaches a certain capacity
            # Use the gc library to manually clean up memory garbage to prevent memory overflow
            if len(stack_out) >= top:
                del stack_out[:]
                gc.collect()
        except IndexError as e:
            # print(e)
            pass
				
@app.route("/")
def index():
	# return the rendered template
	return render_template("index.html")

@app.route("/video_feed")
def video_feed():
	# return the response generated along with the specific media
	# type (mime type)
	return Response(web_stream(web_feed),
		mimetype = "multipart/x-mixed-replace; boundary=frame")

if __name__ == '__main__':
    # construct the argument parser and parse command line arguments
    ap = argparse.ArgumentParser()
    ap.add_argument("-i", "--ip", type=str, default='0.0.0.0',
        help="ip address of the device")
    ap.add_argument("-o", "--port", type=int, default=8000,
        help="ephemeral port number of the server (1024 to 65535)")
    ap.add_argument("-n", "--source-id", type=int, default=0,
        help="video source, default 0")
    args = vars(ap.parse_args())

	# Shared lists
    camera_feed = Manager().list()
    web_feed = Manager().list()

    # The parent process creates a buffer stack and passes it to each child process:
    pw = Process(target=camera_input, args=(camera_feed, args["source_id"], 100))
    pp = Process(target=video_processing, args=(camera_feed, web_feed, 100))
    # Start the child process pw, read camera:
    pw.start()
    # Start the child process pp, process video:
    pp.start()

	# start the flask app
    app.run(host=args["ip"], port=args["port"], debug=True,
        threaded=True, use_reloader=False)

    # Terminate processes to end:
    pp.terminate()
    pw.terminate()